-module(rocket_game).
-include("../bp.hrl").
-include("../bp_ext.hrl").
-export([start/0]).

% Translate external mandatory events to internal blockable ones

translate() ->
	E = case bp:bSync(#rwb{wait=[user_left, user_right, user_up]}) of
			user_left -> left;
			user_right -> right;
			user_up -> up
		end,
	bp:bSync(#rwb{request=[E], wait=?ALL}),
	translate().

% Move rocket to the left
go_left() ->
	bp:bSync(#rwb{wait=[left]}),
	rocket:left(),
	go_left().

% Move rocket to the right
go_right() ->
	bp:bSync(#rwb{wait=[right]}),
	rocket:right(),
	go_right().

% Move rocket down
go_down() ->
	bp:bSync(#rwb{wait=[down]}),
	rocket:down(),
	go_down().

% Push a tick to the external queue every N milliseconds
ticker(N) ->
	receive 
		shutdown -> ok
	after N -> 
			bp_ext:push(tick),
			ticker(N)
	end.

% On tick move try to move rocket down
on_tick() ->
	bp:bSync(#rwb{wait=[tick]}),
	bp:bSync(#rwb{request=[down], wait=?ALL}),
	on_tick().

% Suspend fall for one tick if user presses up
go_up() ->
	bp:bSync(#rwb{wait=[up]}),
	bp:bSync(#rwb{wait=[tick]}),
	bp:bSync(#rwb{wait=[tick], block=[down]}),
	go_up().

% Prevent rocket from moving below the ground
on_ground() ->
	bp:bSync(#rwb{wait=[down]}),
	{_, Y} = rocket:pos(),
	if 
		Y == 0 -> 
			bp:bSync(#rwb{wait=[up], block=[down]});
		true ->
			ok
	end,
	on_ground().

% Prevent rocket from moving left of the board
on_left_bound(X) ->
	if
		X == 0 ->
			bp:bSync(#rwb{wait=[right], block=[left]}),
			on_left_bound(1);
		true ->
			case bp:bSync(#rwb{wait=[left, right]}) of
				left  -> on_left_bound(X-1);
				right -> on_left_bound(X+1)
			end
	end.

% Prevent rocket from moving right of the board
on_right_bound(X, M) ->
	if
		X == M ->
			bp:bSync(#rwb{wait=[left], block=[right]}),
			on_right_bound(X-1, M);
		true ->
			case bp:bSync(#rwb{wait=[left, right]}) of
				left  -> on_right_bound(X-1, M);
				right -> on_right_bound(X+1, M)
			end
	end.

% Prevent rocket from moving left or right more than once per tick
block_mult_moves() ->
	block_mult_moves([]).
block_mult_moves(Block) ->
	Moves = [left, right],
	case bp:bSync(#rwb{wait=[tick | Moves], block=Block}) of
		left ->
			block_mult_moves(Moves);
		right ->
			block_mult_moves(Moves);
		tick ->
			block_mult_moves([])
	end.

% Try to move landing pad randomly each tick
ctrl_landing_pad() ->
	E = element(random:uniform(3), {[pad_left], [pad_right], []}),
	R = bp:bSync(#rwb{wait=[tick], request=E}),
	case R of
		tick -> ctrl_landing_pad();
		_ -> bp:bSync(#rwb{wait=[tick]}), ctrl_landing_pad()
	end.

% Move pad left
on_pad_left() ->
	bp:bSync(#rwb{wait=[pad_left]}),
	landing_pad:left(),
	on_pad_left().
% Move pad right
on_pad_right() ->
	bp:bSync(#rwb{wait=[pad_right]}),
	landing_pad:right(),
	on_pad_right().

% Prevent landing pad from moving left of the board
on_pad_left_bound() ->
	X = landing_pad:pos(),
	if
		X == 0 ->
			bp:bSync(#rwb{wait=[pad_right], block=[pad_left]});
		true ->
			bp:bSync(#rwb{wait=[pad_left]})
	end,
	on_pad_left_bound().

% Prevent landing pad from moving right of the board
on_pad_right_bound(M) ->
	X = landing_pad:pos(),
	if
		X == M ->
			bp:bSync(#rwb{wait=[pad_left], block=[pad_right]});
		true ->
			bp:bSync(#rwb{wait=[pad_right]})
	end,
	on_pad_right_bound(M).

% Test positions for winning (rocket on top of landing pad)
detect_win_lose(P, X, Y) ->
	case bp:bSync(#rwb{wait=[pad_left, pad_right, left, right, down]}) of
		pad_left -> detect_win_lose(P-1, X, Y);
		pad_right -> detect_win_lose(P+1, X, Y);
		left -> detect_win_lose(P, X-1, Y);
		right -> detect_win_lose(P, X+1, Y);
		down ->
			Y1 = Y-1,
			if
				Y1 == 1 andalso X == P ->
					bp:bSync(#rwb{request=[win]});
				Y1 == 0 ->
					bp:bSync(#rwb{request=[lose]});
				true -> 
					detect_win_lose(P, X, Y1)
			end
	end.

% Announce win, and prevent rocket from moving down when on the landing pad
on_win() ->
	bp:bSync(#rwb{wait=[win]}),
	io:format("== WIN ==~n", []).

% Announce lose
on_lose() ->
	bp:bSync(#rwb{wait=[lose]}),
	io:format("== LOSE ==~n", []).

on_game_over() ->
	bp:bSync(#rwb{wait=[win, lose]}),
	game_over().
game_over() ->
	bp:bSync(#rwb{wait=?ALL, block=[down, pad_left, pad_right]}),
	game_over().

echo() ->
	E = bp:bSync(#rwb{wait=?ALL}),
	io:format("~p~n", [E]),
	echo().
						
% Configure and run program
start() ->
	code:add_pathsa([".."]),
	{W, H} = {7, 10},
	I = round(W / 2),
	bp:init(),
	rocket:start(I, H),
	landing_pad:start(I),
	bp_ext:start_idler(),
	bp_ext:start_queue(),
	rocket_panel:start(),
	bp:add(fun translate/0, 1000),
	bp:add(fun go_left/0,   1001),
	bp:add(fun go_right/0,  1010),
	bp:add(fun go_down/0,   1020),
	bp:add(fun go_up/0,     1030),
	bp:add(fun on_tick/0,        1040),
	bp:add(fun on_ground/0,      1050),
	bp:add(fun() -> on_left_bound(I) end, 1060),
	bp:add(fun() -> on_right_bound(I, W-1) end, 1070),
	bp:add(fun block_mult_moves/0, 1080), % comment out this
	bp:add(fun on_pad_left/0,   1090),
	bp:add(fun on_pad_right/0,  1100),
	bp:add(fun on_pad_left_bound/0,  1110),
	bp:add(fun() -> on_pad_right_bound(W-1) end, 1120),
	bp:add(fun ctrl_landing_pad/0, 1135),
	bp:add(fun() -> detect_win_lose(I, I, H) end, 900),
	bp:add(fun on_win/0,           1140),
	bp:add(fun on_lose/0,          1150),
	bp:add(fun on_game_over/0,     1160),
	bp:add(fun echo/0,          1170),
	bp:attach(spawn(fun() -> ticker(2000) end)),
	bp:start().
	
